# DESCRIPTION
# abc2harp is a Perl script that adds harmonica (harp) tab to ABC music
# notation in the form of lyrics text. The key of the harp can be specified
# (default is the same key as the music i.e. "straight harp"). Overbends can
# be allowed or disallowed (default).
#
# Currently only the 10-hole diatonic harp is supported.

# FILENAME
# abc2harp.pl

# AUTHOR
# Gek S. Low

# INSTRUCTIONS

$ver = "0.1";

$help_text = "
abc2harp.pl version $ver.
Copyright (c) Gek S. Low, 2006.
Usage:
    abc2mma.pl <abcfile.abc | abcfile.txt> [options]
	 
    - default output filename is <abcfile>_harp.abc

Options:

-o or -output=<tabfile>
    - output to <tabfile>.abc
-writefile=<yes|no>
    - generates file (overrides -outfile) (default=yes)
-v or verbose
    - turns on verbose mode
-harpkey=<harpkey>
    - sets all harp keys to <harpkey>. Ignores all 
      \"X-ABC2HARP HarpKey\" commands in input file.
-overbends=<yes|no>
    - specifies whether overbends (blows and draws) are allowed
      (default=no). Overrides all \"X-ABC2HARP OverBends\"
      commands in input file.\n";


# NOTES / DISCLAIMERS
# - Only 10-hole diatonics are supported currently
#
# - Inline key changes are not supported, but clef changes should be okay
# 
# - This is NOT a parser! I'm just using mostly string substitution to get
# the results I want. I'm too lazy to write a full abc parser.
#
# - I do not claim to handle all possible abc notation scenarios. There are
# features of abc notation which I don't use, so I'm sure there are many
# cases which are broken.
#
# - Finally, use at your own risk! I'm not responsible for any damages, blah
# blah... that may result from the use of this software.


# TO DO LIST
# - Support for chromatic harps
# - Support for SuperTab font

# CHANGE LOG
# Version 0.1: 15 April 2006

# CONTACT
# gek<remove this>siong AT <remove this - no spam!>yahoo DOT com
# This file and contents is copyright (c) Gek S. Low 2006


# Notes to numbers

sub getNote () {
  my $n = shift;
  
  $n =~ s/\/.*//;
  $n =~ s/^([A-G])#/\^$1/;
  $n =~ s/^([A-G])b/_$1/;
 
  # C,=1, C=13, c=25, c'=37, etc
  $n =~ s/C/\+13/g;
  $n =~ s/D/\+15/g;
  $n =~ s/E/\+17/g;
  $n =~ s/F/\+18/g;
  $n =~ s/G/\+20/g;
  $n =~ s/A/\+22/g;
  $n =~ s/B/\+24/g;
  $n =~ s/c/\+25/g;
  $n =~ s/d/\+27/g;
  $n =~ s/e/\+29/g;
  $n =~ s/f/\+30/g;
  $n =~ s/g/\+32/g;
  $n =~ s/a/\+34/g;
  $n =~ s/b/\+36/g;
  
  $n =~ s/\^/\+1/g;	# sharped notes
  $n =~ s/_/\-1/g;	# flatted notes
  # $n =~ s/\=//;
  $n =~ s/\,/\-12/g;	# one octave below
  $n =~ s/'/\+12/g;	# one octave above
  
  $n = eval($n);
  return $n;
}

# Get names of note numbers based on C=1, D=3,...,B=11
sub getNoteName() {
  my $n = shift;
  my $ret;
  while ($n > 12) { $n = $n - 12; }
  if ($n == 1) { $ret = "C"; }
  if ($n == 2) { $ret = "C#/Db"; }
  if ($n == 3) { $ret = "D"; }
  if ($n == 4) { $ret = "D#/Eb"; }
  if ($n == 5) { $ret = "E"; }
  if ($n == 6) { $ret = "F"; }
  if ($n == 7) { $ret = "F#/Gb"; }
  if ($n == 8) { $ret = "G"; }
  if ($n == 9) { $ret = "G#/Ab"; }
  if ($n == 10) { $ret = "A"; }
  if ($n == 11) { $ret = "A#/Bb"; }
  if ($n == 12) { $ret = "B"; }
  return $ret;
}

# Get relative major of the minor key
sub min2maj() {
  my $key = shift;
  my $k;
  # Is this a minor key?
  if ($key =~ /m/) {
	 $key =~ s/m(inor)?//;
	 $k = &getNote($key);
	 $k = $k + 3;	# convert to relative major
	 if ($k > 24) { $k = $k - 12; }
  } else {
	 $k = &getNote($key, $k);
  }
  return $k;
}

# Fixes the abc notes by applying the flats and sharps accordingly

sub getTrueNote() {
  my $note = shift;
  my $key = shift;

  # First, check if it's a "natural" note. If so, return it
  if ($note =~ /\=/) {
	 $note =~ s/\=//g;
	 $n = &getNote($note);
	 return $n;
  }
   
  # Okay, let's look at the key sig
  # Sharps
  $n = &getNote($note);
  if ($note =~ /__/) { return $n; }	# double flats
  if ($note =~ /\^\^/) { return $n; }	# double sharps

  $k = &min2maj($key);
  if ($k == 13) { return $n; }
  if ($n == 6 || $n == 18 || $n == 30) { $n++; } # F -> F#
  if ($k == 20) { return $n; }
  if ($n == 1 || $n == 13 || $n == 25 || $n == 37) { $n++; } # C -> C#
  if ($k == 15) { return $n; }
  if ($n == 8 || $n == 20 || $n == 32) { $n++; } # G -> G#
  if ($k == 22) { return $n; }
  if ($n == 3 || $n == 15 || $n == 27) { $n++; } # D -> D#
  if ($k == 17) { return $n; }
  if ($n == 10 || $n == 22 || $n == 34) { $n++; } # A -> A#
  if ($k == 24) { return $n; }
  if ($n == 5 || $n == 17 || $n == 29) { $n++; } # E -> F
  if ($k == 19) { return $n; }
  if ($n == 12 || $n == 24 || $n == 36) { $n++; } # B -> C
  if ($k == 14) { return $n; }
  
  # Flats
  $n = &getNote($note);
  $k = &min2maj($key);
  if ($n == 12 || $n == 24 || $n == 36) { $n--; } # B -> Bb
  if ($k == 18) { return $n; }
  if ($n == 5 || $n == 17 || $n == 29) { $n--; } # E -> Eb
  if ($k == 23) { return $n; }
  if ($n == 10 || $n == 22 || $n == 34) { $n--; } # A -> Ab
  if ($k == 16) { return $n; }
  if ($n == 3 || $n == 15 || $n == 27) { $n--; } # D -> Db
  if ($k == 21) { return $n; }
  if ($n == 8 || $n == 20 || $n == 32) { $n--; } # G -> Gb
  if ($k == 14) { return $n; }
  if ($n == 1 || $n == 13 || $n == 25 || $n == 37) { $n--; } # C -> Cb
  if ($k == 19) { return $n; }
  if ($n == 6 || $n == 18 || $n == 30) { $n--; } # F -> E
  if ($k == 24) { return $n; }

  # oops, shouldn't reach this place
  print "There was an error recognising the notes. The tab will probably be wrong.\n";
  return 0;
}

sub setHarpKey() {
  my $harpkey = shift;
  $transposeharp = &getNote($harpkey) - 12;
  if ($transposeharp >= 8 && $transposeharp <=12) {
	 $transposeharp = 13 - $transposeharp;
  } else {
	 $transposeharp = 1 - $transposeharp;
  }
  #print $transposeharp."\n";
}

sub filterOverBends() {
  my $n = shift;
  my $ob = shift;

  if ($ob ne "yes") {
	 if ($n == 4 || $n == 16 || $n == 19 || $n == 23 || $n == 26 || $n == 31
		|| $n == 33 || $n == 38) {
		$n = 0;
	 }
  }
  return $n;
}



# harptab layout (C harp)
$tab[0] = "x"; # unknown

$tab[1] = "~1"; # C
$tab[2] = "\\-1'"; # C# / Db
$tab[3] = "\\-1"; # D
$tab[4] = "~1o"; # E# / Eb
$tab[5] = "~2"; # E
$tab[6] = "\\-2''"; # F
$tab[7] = "\\-2'"; # F# / Gb
$tab[8] = "\\-2"; # G
$tab[9] = "\\-3'''"; # G# / Ab
$tab[10] = "\\-3''"; # A
$tab[11] = "\\-3'"; # A# / Bb
$tab[12] = "\\-3"; # B

$tab[13] = "~4"; # C
$tab[14] = "\\-4'"; # C# / Db
$tab[15] = "\\-4"; # D
$tab[16] = "~4o"; # E# / Eb
$tab[17] = "~5"; # E
$tab[18] = "\\-5"; # F
$tab[19] = "~5o"; # F# / Gb
$tab[20] = "~6"; # G
$tab[21] = "\\-6'"; # G# / Ab
$tab[22] = "\\-6"; # A
$tab[23] = "~6o"; # A# / Bb
$tab[24] = "\\-7"; # B

$tab[25] = "~7"; # C
$tab[26] = "\\-7o"; # C# / Db
$tab[27] = "\\-8"; # D
$tab[28] = "~8'"; # E# / Eb
$tab[29] = "~8"; # E
$tab[30] = "\\-9"; # F
$tab[31] = "~9o"; # F# / Gb
$tab[32] = "~9"; # G
$tab[33] = "\\-9o"; # G# / Ab
$tab[34] = "\\-10"; # A
$tab[35] = "~10''"; # A# / Bb
$tab[36] = "~10'"; # B

$tab[37] = "~10'"; # B
$tab[38] = "~10o'"; # B

# Set up defaults
$overbends = "no";
$keysig = "C";
$harpkey = "";
$arg_harpkey = "";
$arg_overbends = "";
$octaveadjust = 0;

$writefile = "yes";
$verbose = 0;

$abcfile = $ARGV[0];
$tabfile = $abcfile;
$tabfile =~ s/(\.\w\w\w)$/_harp$1/;

if ($ARGV[0] eq "") {
  print $help_text;
  exit(0);
}

foreach $arg (@ARGV) {
  if ($arg =~ /\-o(utput)?\=(.*)$/) {
	 $tabfile = $2;
  }
  if ($arg =~ /\-writefile\=(.*)$/) {
	 $writefile = $1;
  }
  if ($arg =~ /\-v(erbose)?$/) {
	 $verbose = 1;
  }
  if ($arg =~ /\-harpkey\=(.*)$/) {
	 $arg_harpkey = $1;
	 print "Harp key $arg_harpkey specified. Will override all \"X-ABC2HARP HarpKey\" commands.\n\n";
  }
  if ($arg =~ /\-overbends\=(.*)$/) {
	 $arg_overbends = $1;
	 $overbends = $arg_overbends;
  }
}

open(INPUT, $abcfile) or die "Unable to open $abcfile.\n";

@input = <INPUT>;

close(INPUT);

if ($writefile eq "yes") {
  open (OUTPUT, ">$tabfile") or die "Unable to create $tabfile.\n";
} else {
  *OUTPUT = *STDOUT;
}

$line_count=0;
$abcheaderdone=0;

foreach $line (@input) {
  $line_count++;
  if (($line =~ /^\w:/) or ($line =~ /^%/) or ($line eq "\n")) {
	 print OUTPUT $line;
	 
	 if ($line =~ /^X:\s*(.*)$/) {
		# new song, can reset defaults
		$abcheaderdone = 0;
		
		$keysig = "C";
		if ($arg_harpkey ne "") {
		  $harpkey = $arg_harpkey;
		} else {
		  $harpkey = "";
		}
		$transposeharp = 0;
		$octaveadjust = 0;
		if ($arg_overbends ne "") {
		  $overbends = $arg_overbends;
		} else {
		  $overbends = "no";
		}
		if ($verbose == 1) { print "\nSong number: $1\n"; }
	 }
	 # print the title
	 if ($line =~ /^T:\s*(.*)$/) {
		if ($verbose == 1) { print "Title: $1\n"; }
	 }
	 # this is an ABC field, comment or command, so just print it
	 if ($line =~ /^K:\s*(.*)$/) {
		$abcheaderdone = 1;
		
		$octaveadjust = 0;
		$keysig = $1;	# note for future: need to adjust for treble+-8
		if ($keysig =~ /treble\+8/) { $octaveadjust = 12; }
		if ($keysig =~ /treble\-8/) { $octaveadjust = -12; }
		$keysig =~ s/\s?(treble[+-]?\d?|bass\d?|alto\d?|none|perc)//;
		$keysig =~ s/\s?major//;
		if ($verbose == 1) { print "Key signature: ".$keysig."\n"; }
		$keysig =~ s/^([A-G])#/\^$1/;
		$keysig =~ s/^([A-G])b/_$1/;
		
		# has a harp key been set?
		if ($harpkey eq "") {
		  # default harp key is the (relative major) key signature
		  $harpkey = &getNoteName(&min2maj($keysig));
		  &setHarpKey($harpkey);
		  if ($verbose == 1) { print "Harp key: $harpkey\n"; }
		}
		print OUTPUT "\%\%text $harpkey harp\n";

		# print the overbends info
		if ($verbose == 1) { print "Overbends: $overbends\n"; }
	 }
	 # abc2harp harp key command
	 if ($line =~ /^%%X\-ABC2HARP HarpKey (.*)$/) {
		if ($arg_harpkey eq "") {
		  $harpkey = $1;
		  &setHarpKey($harpkey);
		  if ($verbose == 1) { print "Harp key: $harpkey\n"; }
		  if ($abcheaderdone == 1) { print OUTPUT "\%\%text $harpkey harp\n"; }
		}
	 }
	 # abc2harp overbends command
	 if ($line =~ /^%%X\-ABC2HARP OverBends (.*)$/) {
		if ($arg_overbends eq "") {
		  $overbends = $1;
		  if ($verbose == 1) { print "Overbends: $overbends\n"; }
		} else {
		  $overbends = $arg_overbends;
		}
	 }
  } else {
	 # this should be a musical line
	 print OUTPUT $line;
	 
	 # make a copy - this will be the harp tab line
	 $harpline = $line;

	 $harpline =~ s/\[\w:.*?\]//g;
	 
	 # strip to individual notes
	 $harpline =~ s/%.*?$//;	# remove trailing comments
	 $harpline =~ s/\\//g;		# remove continuation lines
	 
	 $harpline =~ s/\"(.*?)\"//g;	# remove all strings
	 $harpline =~ s/[<>\.~]//g;	# remove - < > . ~
	 $harpline =~ s/!.*?!//g;		# remove !symbols!
	 $harpline =~ s/{.*?}//g;		# remove graces {}
	 $harpline =~ s/\(([^\(]*?)\)/$1/g;		# "de"-slur
	 $harpline =~ s/\[([\^_=]*?[A-Ga-gzx][\,']*).*?\]/$1/g;	# "de"-chord
	 $harpline =~ s/\(\d//g;		# "de"-tuplet
	 $harpline =~ s/[\|\]:]//g;	# remove barlines
	 $harpline =~ s/\[\d//g;		# remove numbered repeats

	 $harpline =~ s/([\^_=]*?[A-Ga-gzx][\,']*)\d?\/*?/$1 /g;

	 $harpline =~ s/\-\s*[\^_=]*?[A-Ga-gzx][\,']*/\*/g;	# ties will be skipped
	 $harpline =~ s/\-//g;	# clean up extra ties
	 
	 $harpline =~ s/[zx]\d?//g;		# remove rests

	 # convert to tab
	 $harpline =~ s/\s+/ /g;
	 @notes = split(" ", $harpline);
	 foreach $note (@notes) {
		if ($note ne "*") {
		  $abcnote = $note;
		  $note_index = &getTrueNote($abcnote, $keysig) + $octaveadjust + $transposeharp;
		  if ($note_index < 0) { $note_index = 0; }
		  if ($note_index > 38) { $note_index = 0; }
		  $note_index = &filterOverBends($note_index, $overbends);
		  $note = $tab[$note_index];
		  if ($note eq "x") { print "Line $line_count: Warning: $abcnote is not playable on the $harpkey harp\n"; }
		}
	 }

	 $harpline = join(' ', @notes);
	 
	 $harpline = "w: ".$harpline;
	 $harpline =~ s/\s+/ /g;		# remove extra spaces
	 
	 if ($harpline ne "w: ") { print OUTPUT "$harpline\n"; }
  }
}

if ($writefile eq "yes") { print "\nCreated $tabfile\n"; }

